import logging
import datetime as dt

from . import valuation
from .validation import check_trades

logger = logging.getLogger(__name__)


def add_portfolio_purchase_price(trades, initial_purchase_price):
    trades["portfolio_purchase_price"] = 0
    trades.loc[trades["trade_side"] == "BUY", "portfolio_purchase_price"] = (
        trades["amount"] + trades["fee"]
    )
    trades["portfolio_purchase_price"] = trades["portfolio_purchase_price"].cumsum()
    trades["portfolio_purchase_price"] += initial_purchase_price


def filter_sales_add_portfolio_value(trades, initial_portfolio):
    portfolio = valuation.valuate_portfolio(trades, initial_portfolio)
    print("Valuate done")
    value = portfolio["value", "TOTAL"]
    value = value.rename("portfolio_value").to_frame()
    sales = trades.join(value, how="inner")
    return sales


def compute_purchase_price_fraction(
    amount, value, purchase_price, purchase_price_net, fraction, fraction_sum
):
    frac_sum = 0
    for i in range(0, len(amount)):
        fraction_sum[i] = frac_sum
        # Decrease the portfolio purchase price by the sum of the fractions
        # of pruchase price aldready sold
        purchase_price_net[i] = purchase_price[i] - fraction_sum[i]
        # The percentage of the portfolio that was sold by this transaction
        percentage_sold = amount[i] / value[i]
        # Compute the fraction of the purchase price that was sold
        # by this transaction
        fraction[i] = purchase_price_net[i] * percentage_sold
        frac_sum += fraction[i]


def compute_taxable_pnls_detailed(
    trades, initial_portfolio=None, initial_purchase_price=0.0
):
    """Computes your taxable PnL for each sale in the trades DataFrame

    In the french tax law (CGI art. 150 VH), taxable events are sales of
    crypto-currencies for official (fiat) currencies such as Euro. The PnL
    (profit and loss) arising from each sale must be reported in the form 2086
    (Formulaire n°2086).

    This function computes the PnL to be reported for each sale on form 2086, and
    other information that also needs to be reported on form 2086. It returns
    a DataFrame such as::

                     datetime trade_side cryptocurrency  quantity    amount       fee
        2 2020-09-05 16:50:00       SELL            BTC       0.5   4361.35  21.80675
        3 2020-09-08 12:40:00       SELL            ETH       5.0   1425.35   7.12675
        6 2020-12-21 09:30:00       SELL            BTC       1.0  19531.69  97.65845

            amount_net  portfolio_value  portfolio_purchase_price
        2   4339.54325        10226.900                11286.4716
        3   1418.22325         5679.970                11286.4716
        6  19434.03155        31934.785                22507.6584

        purchase_price_fraction  purchase_price_fraction_sum
        2              4813.213477                     0.000000
        3              1624.420281                  4813.213477
        6              9828.616024                  6437.633759

           portfolio_purchase_price_net          pnl
        2                  11286.471600  -473.670227
        3                   6473.258123  -206.197031
        6                  16070.024641  9605.415526

    The meaning of each column is as follows:

    .. |reportable| replace:: **[!]** This information has to be reported on form 2086
        (see :py:func:`coin2086.compute_taxable_pnls` to known in which cell).

    datetime
        The date and time of your trade, copied from the input ``trades`` DataFrame.

        |reportable|
    trade_side
        The side (BUY or SELL) of your trade, copied from the input ``trades``
        DataFrame
    cryptocurrency
        The crypto currency you sold, copied from the input ``trades`` DataFrame
    quantity
        The quantity of crypto-currency you sold, copied from the input ``trades``
        DataFrame
    amount
        The amount in euro you received in exchange for you sale of
        crypto-currencies, copied from the input ``trades`` DataFrame.

        |reportable|
    fee
        The fees in euro you paid for your trade, copied from the input trades
        DataFrame.

        |reportable|
    amount_net
        The amount, net of fees, your received in exchange for your sale
        of crypto-currencies. This is amount - fee.
    portfolio_value
        The total value of your portfolio of crypto-currencies before the sale,
        as computed by :py:func:`coin2086.valuate_portfolio`

        |reportable|
    portfolio_purchase_price
        The total purchase price of your porfolio, net of fees, but not
        accouting for sales.

        |reportable|
    purchase_price_fraction
        The fraction of the purchase price that was sold (Fractions de capital
        initial, see Notice du formulaire n°2086). For instance, if : (1) your
        portfolio is worth 10226.90 Euros, (2) you receive 4339.54 Euros for a
        crypto-currency sale and (3) the total purchase price for your portfolio
        was 11286.4716, it will be considered that of have sold
        4339.54 / 10226.90 = 42.6458% of your portfolio. The purchase price for
        these 42.43% of your porfolio (purchase_price_fraction) is 42.6458%
        of 11286.4716 = 4813.213477.
    purchase_price_fraction_sum
        The cumulative sum of the purchase_price_fraction column

        |reportable|
    portfolio_purchase_price_net
        The total purchase price of your porfolio, net of fees, and net of the
        cumulative sum of the purchase price fractions. This is
        portfolio_purchase_price - purchase_price_fraction_sum.

        |reportable|
    pnl
        The PnL (profit and loss) for this sale. You will pay 30% of the
        ``sum()`` of this column as taxes, only if this is a positive number (no
        taxes if this is a negative number)

        |reportable|

    .. note::
        Note that the indexes of this DataFrame are the indexes of the sales
        (trades with trade_side = SELL) of your ``trades`` DataFrame.

    The input trades DataFrame must have one line per trade, as follows::

                     datetime trade_side cryptocurrency  quantity     price
        0 2020-07-28 10:20:00        BUY            BTC       1.0   9262.42
        1 2020-09-01 12:20:00        BUY            ETH       5.0    393.58
        2 2020-09-05 16:50:00       SELL            BTC       0.5   8722.70
        3 2020-09-08 12:40:00       SELL            ETH       5.0    285.07
        4 2020-09-16 17:10:00        BUY            BTC       1.0   9247.51
        5 2020-11-07 15:40:00        BUY            ETH       5.0    383.57
        6 2020-12-21 09:30:00       SELL            BTC       1.0  19531.69

        base_currency    amount       fee
        0           EUR   9262.42  46.31210
        1           EUR   1967.90   9.83950
        2           EUR   4361.35  21.80675
        3           EUR   1425.35   7.12675
        4           EUR   9247.51  46.23755
        5           EUR   1917.85   9.58925
        6           EUR  19531.69  97.65845

    .. include:: ../../docs/includes/input_columns.rst

    Args:
        trades (pandas.DataFrame): .. include:: ../../docs/includes/arg_trades.rst
        initial_portfolio (dict):  .. include:: ../../docs/includes/arg_initial_portfolio.rst
        initial_purchase_price (float): The purchase price of the initial_portfolio

    Returns:
        pandas.DataFrame: The DataFrame containing the information to be reported
        on form 2086 for each sale in the trades DataFrame.
    """

    check_trades(trades)
    trades = trades.copy()
    add_portfolio_purchase_price(trades, initial_purchase_price)
    sales = filter_sales_add_portfolio_value(trades, initial_portfolio)
    sales = sales[
        [
            "datetime",
            "trade_side",
            "cryptocurrency",
            "quantity",
            "amount",
            "fee",
            "portfolio_purchase_price",
            "portfolio_value",
        ]
    ]
    sales["portfolio_purchase_price_net"] = 0.0
    sales["purchase_price_fraction_sum"] = 0.0
    sales["purchase_price_fraction"] = 0.0
    compute_purchase_price_fraction(
        sales["amount"].values,
        sales["portfolio_value"].values,
        sales["portfolio_purchase_price"].values,
        sales["portfolio_purchase_price_net"].values,
        sales["purchase_price_fraction"].values,
        sales["purchase_price_fraction_sum"].values,
    )
    sales["amount_net"] = sales["amount"] - sales["fee"]
    sales["pnl"] = sales["amount_net"] - sales["purchase_price_fraction"]
    sales = sales[
        [
            "datetime",
            "trade_side",
            "cryptocurrency",
            "quantity",
            "amount",
            "fee",
            "amount_net",
            "portfolio_value",
            "portfolio_purchase_price",
            "purchase_price_fraction",
            "purchase_price_fraction_sum",
            "portfolio_purchase_price_net",
            "pnl",
        ]
    ]
    return sales


def compute_taxable_pnls(
    trades, year, initial_portfolio=None, initial_purchase_price=0.0
):
    """
    Computes your taxable PnL for each sale in the trades DataFrame

    In the french tax law (CGI art. 150 VH), taxable events are sales of
    crypto-currencies for official (fiat) currencies such as Euro. The PnL
    (profit and loss) arising from each sale must be reported in the form 2086
    (Formulaire n°2086).

    This function computes the PnL to be reported for each sale on form 2086, and
    other information that also needs to be reported on form 2086. It returns
    a DataFrame such as::

              Description Date de la cession [datetime]
        2  SELL 0.50 BTC           2020-09-05 16:50:00
        3  SELL 5.00 ETH           2020-09-08 12:40:00
        6  SELL 1.00 BTC           2020-12-21 09:30:00

           Valeur globale portefeuille [portfolio_value]  Prix de cession [amount]
        2                                      10226.900                   4361.35
        3                                       5679.970                   1425.35
        6                                      31934.785                  19531.69

           Frais de cession [fee]  Prix de cession net des frais [amount_net]
        2                21.80675                                  4339.54325
        3                 7.12675                                  1418.22325
        6                97.65845                                 19434.03155

          Prix total d'acquistion [portfolio_purchase_price]
        2                                         11286.4716
        3                                         11286.4716
        6                                         22507.6584

        Fractions de capital initial [purchase_price_fraction_sum]
        2                                           0.000000
        3                                        4813.213477
        6                                        6437.633759

        Prix total d'acquistion net [portfolio_purchase_price_net]
        2                                       11286.471600
        3                                        6473.258123
        6                                       16070.024641

           Plus-values et moins-values [pnl]
        2                        -473.670227
        3                        -206.197031
        6                        9605.415526

    You may copy directly the value of each cell (Date de la cession,
    Valeur globale portefeuille etc.) direcly in the form 2086
    (Formulaire n°2086). Internally, this

    The input trades DataFrame must have one line per trade, as follows::

                     datetime trade_side cryptocurrency  quantity     price
        0 2020-07-28 10:20:00        BUY            BTC       1.0   9262.42
        1 2020-09-01 12:20:00        BUY            ETH       5.0    393.58
        2 2020-09-05 16:50:00       SELL            BTC       0.5   8722.70
        3 2020-09-08 12:40:00       SELL            ETH       5.0    285.07
        4 2020-09-16 17:10:00        BUY            BTC       1.0   9247.51
        5 2020-11-07 15:40:00        BUY            ETH       5.0    383.57
        6 2020-12-21 09:30:00       SELL            BTC       1.0  19531.69

        base_currency    amount       fee
        0           EUR   9262.42  46.31210
        1           EUR   1967.90   9.83950
        2           EUR   4361.35  21.80675
        3           EUR   1425.35   7.12675
        4           EUR   9247.51  46.23755
        5           EUR   1917.85   9.58925
        6           EUR  19531.69  97.65845

    .. include:: ../../docs/includes/input_columns.rst

    Args:
        trades (pandas.DataFrame): .. include:: ../../docs/includes/arg_trades.rst
        year (int): The year to report your trades for. Usually, this is the
            last year.
        initial_portfolio (dict): .. include:: ../../docs/includes/arg_initial_portfolio.rst
        initial_purchase_price (float): The purchase price of the initial_portfolio

    Returns:
        (pandas.DataFrame, float): The DataFrame containing the information
        to be reported on form 2086 for each sale in the ``trades`` DataFrame,
        with the sum of the PnLs (Plus et moins values)
    """
    check_trades(trades)
    sales = compute_taxable_pnls_detailed(
        trades, initial_portfolio, initial_purchase_price
    )
    start_date = dt.datetime.combine(dt.date(year, 1, 1), dt.time.min)
    end_date = dt.datetime.combine(dt.date(year, 12, 31), dt.time.max)
    sales = sales[(sales["datetime"] >= start_date) & (sales["datetime"] <= end_date)]
    total_pnl = sales["pnl"].sum()
    sales["Description"] = sales["description"] = (
        sales["trade_side"]
        + " "
        + sales["quantity"].map("{:.2f}".format)
        + " "
        + sales["cryptocurrency"]
    )
    names = {
        "datetime": "Date de la cession [datetime]",
        "portfolio_value": "Valeur globale portefeuille [portfolio_value]",
        "amount": "Prix de cession [amount]",
        "fee": "Frais de cession [fee]",
        "amount_net": "Prix de cession net des frais [amount_net]",
        "portfolio_purchase_price": "Prix total d'acquistion [portfolio_purchase_price]",
        "purchase_price_fraction_sum": "Fractions de capital initial [purchase_price_fraction_sum]",
        "portfolio_purchase_price_net": "Prix total d'acquistion net [portfolio_purchase_price_net]",
        "pnl": "Plus-values et moins-values [pnl]",
    }
    columns = ["Description"] + list(names.keys())
    sales = sales[columns]
    sales = sales.rename(columns=names)
    return sales, total_pnl



import logging
import collections
import datetime as dt

import requests
import pandas as pd

from . import pricedownload
from .validation import check_trades


logger = logging.getLogger(__name__)


def valuate_portfolio(trades, initial_portfolio=None):
    """Determines the valuation of the porfolio before each sale

    The formula used to compute your taxable PnL (profit and loss) from each
    sale requires determining the total value of your portfolio of
    crypto-currencies before each sale. This is the purpose of this function.
    This function retrieves the prices of crypto-currencies from the public
    Bitstamp API to determine the value of your portfolio. Even if you did not
    trade on Bistamp, the prices should be close enough for our purposes.
    This function returns a column multi-indexed DataFrame. The first level
    of the column index is either quantity, sell_price, public_price, ref_price
    and value. The second level of the index are the crypto-currencies. For
    instance::

                       quantity      sell_price         public_price
        cryptocurrency      BTC  ETH        BTC     ETH          BTC     ETH
        2                   1.0  5.0    8722.70     NaN      8722.70  300.84
        3                   0.5  5.0        NaN  285.07      8509.24  285.07
        6                   1.5  5.0   19531.69     NaN     19531.69  527.45

                       ref_price              value
        cryptocurrency       BTC     ETH        BTC      ETH      TOTAL
        2                8722.70  300.84   8722.700  1504.20  10226.900
        3                8509.24  285.07   4254.620  1425.35   5679.970
        6               19531.69  527.45  29297.535  2637.25  31934.785

    The meaning of each column is as follows:

    quantity
        The quantity of each crypto-currency in the porfolio
    sell_price
        The price you sold a crypto-currency at. It will be a non-null value for
        the crypto-currency you sold, and NaN for the other ones
    public_price
        The known price on Bitstamp of each crypto-currency at the moment
        of your sale trade (the close price of the minute you traded is used)
    ref_public
        The union of sell_price and public_price. For a given crypto-currency,
        if the sale price is known, it is used, otherwise public_price is used
    value
        The value of each of your crypto-currencies holdings. This is computed
        by multiplying quantity by ref_price. This column has an additional
        TOTAL sub-column that gives the total value of your portfolio

    Note that the indexes of this DataFrame are the indexes of the sales
    (trades with trade_side = SELL) of your trades DataFrame.

    Args:
        trades (pandas.DataFrame): .. include:: ../../docs/includes/arg_trades.rst
        initial_portfolio (dict):  .. include:: ../../docs/includes/arg_initial_portfolio.rst

    Returns:
        pandas.DataFrame: The DataFrame containing the composition of the
            portfolio, the valuation of the portfolio and the reference
            prices used for valuation before each sale.
    """
    check_trades(trades)
    sales = trades[trades["trade_side"] == "SELL"]
    portfolio = unstack_portfolio_composition(trades, sales, initial_portfolio)
    portfolio = add_sell_prices(portfolio, sales)
    portfolio = add_public_prices(portfolio, sales)
    portfolio = merge_rates_and_valuate(portfolio)
    return portfolio[["quantity", "sell_price", "public_price", "ref_price", "value"]]


def add_initial_portfolio(portfolio, initial_portfolio):
    if initial_portfolio is None:
        return
    for crypto, qty in initial_portfolio.items():
        if crypto not in portfolio["quantity"].columns:
            portfolio["quantity", crypto] = 0.0
        portfolio["quantity", crypto] += qty


def unstack_portfolio_composition(trades, sales, initial_portfolio):
    portfolio = trades[["cryptocurrency", "quantity", "trade_side"]].copy()
    # Compute a signed quantity for each trade by multiplying by
    # -1 for sales and 1 for purchases
    portfolio["sign"] = -1
    portfolio.loc[portfolio["trade_side"] == "BUY", "sign"] = 1
    portfolio["quantity"] = portfolio["quantity"] * portfolio["sign"]
    portfolio = portfolio.drop(columns=["sign", "trade_side"])
    # Unstack the composition of the portofolio after each transaction
    portfolio = portfolio.set_index("cryptocurrency", append=True).unstack().fillna(0)
    # We need the valuation of the portoflio *before* each transaction
    # so shift by one
    portfolio["quantity"] = portfolio["quantity"].cumsum().shift(1, fill_value=0)
    # Add the initial portfolio to the first trade
    add_initial_portfolio(portfolio, initial_portfolio)
    return portfolio.reindex(sales.index)


def merge_rates_and_valuate(portfolio):
    portfolio = portfolio.stack()
    portfolio["ref_price"] = portfolio["sell_price"].fillna(portfolio["public_price"])
    portfolio["value"] = portfolio["ref_price"] * portfolio["quantity"]
    portfolio = portfolio.unstack()
    portfolio["value", "TOTAL"] = portfolio["value"].sum(axis=1)
    return portfolio


def add_sell_prices(portfolio, sales):
    sell_prices = sales[["cryptocurrency", "price"]]
    sell_prices = sell_prices.rename(columns={"price": "sell_price"})
    sell_prices = sell_prices.set_index("cryptocurrency", append=True).unstack()
    return portfolio.join(sell_prices, how="outer")


def add_public_prices(portfolio, sales):
    pricedown = pricedownload.reference_price_downloader()
    dated_portfolio = portfolio["quantity"].join(sales["datetime"])
    price_records = []
    for rec in dated_portfolio.to_dict(orient="records"):
        dtime = rec["datetime"]
        del rec["datetime"]
        price_rec = {}
        for crypto in rec.keys():
            price_rec[crypto] = pricedown.download_price(crypto, dtime.to_pydatetime())
        price_records.append(price_rec)
    public_prices = pd.DataFrame(price_records, index=sales.index)
    public_prices.columns = pd.MultiIndex.from_product(
        [["public_price"], public_prices.columns]
    )
    return portfolio.join(public_prices, how="outer")



import pandas as pd


def split_column_get_value(col):
    return col.str.split().str[0].astype(float)


def split_column_get_unit(col):
    return col.str.split().str[1]


def normalize_bitstamp_transactions(trans, start_date):
    """
    Take a DataFrame of transactions as exported by Bitstamp, and
    return a DataFrame of *normalized* trades, that can be used by the functions
    of the public API of coin2086. Transactions exported from Bitstamp may include
    deposits, withdrawals or other operations that are not used for the
    computation of the PnL. These operations will be filtered out, and only
    crypto-currencies purchase and sales will be kept.

    Args:
        trans (pandas.DataFrame): A DataFrame of transactions, as exported from your
            Bitstamp profile. This is usually exported as csv from Bitstamp, and
            read with `pandas.read_csv()`.
        start_date(timestamp, str or datetime): The start date at which to
            start parsing transations from. Transactions before start_date will
            be ignored. This will be converted into a pandas datetime using
            `pandas.to_datetime()`.

    Returns:
        pandas.DataFrame: A DataFrame of normalized crypto-currency buy and sell
        trades, that can be used by coin2086 public API functions.
    """
    # Select only trade (Market) transaction and ignore Deposits, Withdrawals etc.
    trades = trans[trans["Type"] == "Market"].copy()
    trades["datetime"] = pd.to_datetime(trades["Datetime"])
    trades["quantity"] = split_column_get_value(trades["Amount"])
    trades["cryptocurrency"] = split_column_get_unit(trades["Amount"])
    trades["price"] = split_column_get_value(trades["Rate"])
    trades["rate_unit"] = split_column_get_unit(trades["Rate"])
    trades["amount"] = split_column_get_value(trades["Value"])
    trades["base_currency"] = split_column_get_unit(trades["Value"])
    trades["fee"] = split_column_get_value(trades["Fee"])
    trades["fee_unit"] = split_column_get_unit(trades["Fee"])
    trades["trade_side"] = trades["Sub Type"].str.upper()
    cols = [
        "datetime",
        "trade_side",
        "cryptocurrency",
        "quantity",
        "price",
        "base_currency",
        "amount",
        "fee",
    ]
    trades = trades[trades["datetime"] > pd.to_datetime(start_date)].reset_index()
    return trades[cols]



import collections
import logging
import abc
import threading
import datetime as dt

import requests
import pandas as pd

logger = logging.getLogger(__name__)


TradingPair = collections.namedtuple("TradingPair", ["base", "quote"])


class PriceDownloader(abc.ABC):
    @abc.abstractmethod
    def download_price(self, crypto, dtime):
        pass

    @abc.abstractproperty
    def supported_crypto_list(self):
        pass


class CachedPriceDownloader(PriceDownloader):
    """Base class for crypto-currency price downloader classes.
    This class maintains a cache so that child classes can avoid
    hitting exchange or data providers  APIs too often
    """

    def __init__(self):
        self.cache = collections.defaultdict(dict)

    def _add_price_to_cache(self, crypto, dtime, price):
        crypto_cache = self.cache[crypto]
        crypto_cache[dtime] = price

    def find_price_in_cache(self, crypto, dtime):
        crypto_cache = self.cache[crypto]
        return crypto_cache.get(dtime)


def round_datetime(dtime, interval):
    return pd.to_datetime(dtime).round(interval).to_pydatetime()


FIAT_CURRENCIES = {"USD", "EUR", "CAD", "JPY", "GBP", "CHF", "AUD", "KRW"}


def is_fiat_currency(code):
    return code in FIAT_CURRENCIES


def bitstamp_download_supported_pairs():
    PAIRS_URL = "https://www.bitstamp.net/api/v2/trading-pairs-info/"
    resp = requests.get(PAIRS_URL)
    pairs = []
    for pinfo in resp.json():
        name = pinfo["name"]
        base, quote = name.split("/")
        pairs.append(TradingPair(base=base, quote=quote))
    return pairs


def bitstamp_download_minute_bins(crypto, dtime, limit=100):
    OHLC_URL = "https://www.bitstamp.net/api/v2/ohlc/{pair}/".format(
        pair=crypto.lower() + "eur"
    )
    parameters = {"start": int(dtime.timestamp()), "step": 60, "limit": limit}
    logger.info(f"Downloading prices from {OHLC_URL}, parameters: {parameters}")
    resp = requests.get(OHLC_URL, params=parameters)
    return resp.json()


class BitstampMinuteClosePriceDownloader(CachedPriceDownloader):
    TIME_INTERVAL = "min"

    def __init__(self):
        super().__init__()
        self._supported_crypto_list = self._download_supported_crypto_list()

    @property
    def supported_crypto_list(self):
        return self._supported_crypto_list

    @staticmethod
    def _download_supported_crypto_list():
        pairs = bitstamp_download_supported_pairs()
        return sorted(
            [p.base for p in pairs if p.quote == "EUR" and not is_fiat_currency(p.base)]
        )

    def download_price(self, crypto, dtime):
        dtime = round_datetime(dtime, self.TIME_INTERVAL)
        cached_price = self.find_price_in_cache(crypto, dtime)
        if cached_price is None:
            self._download_price_add_to_cache(crypto, dtime)
        cached_price = self.find_price_in_cache(crypto, dtime)
        if cached_price is None:
            raise RuntimeError(f"Could not download price for {crypto} at {dtime}")
        return cached_price

    def _download_price_add_to_cache(self, crypto, dtime):
        resp = bitstamp_download_minute_bins(crypto, dtime)
        assert resp["data"]["pair"] == crypto.upper() + "/EUR"
        bins = resp["data"]["ohlc"]
        for b in bins:
            dtime = dt.datetime.fromtimestamp(int(b["timestamp"]))
            rounded = round_datetime(dtime, self.TIME_INTERVAL)
            if not rounded == dtime:
                raise ValueError(
                    f"Bitstamp API returned a bin time that "
                    "is not rounded to {self.time_interval}: {dtime}"
                )
            self._add_price_to_cache(crypto, dtime, float(b["close"]))


KRAKEN_TO_USUAL_CODEBOOK = {
    "XDAO": "DAO",
    "XETC": "ETC",
    "XETH": "ETH",
    "XICN": "ICN",
    "XLTC": "LTC",
    "XMLN": "MLN",
    "XNMC": "NMC",
    "XREP": "REP",
    "XREPV2": "REPV2",
    "XXBT": "BTC",
    "XBT": "BTC",
    "XXDG": "DOGE",
    "XDG": "DOGE",
    "XXLM": "XLM",
    "XXMR": "XMR",
    "XXRP": "XRP",
    "XXTZ": "XTZ",
    "XXVN": "XVN",
    "XZEC": "ZEC",
    "ZAUD": "AUD",
    "ZCAD": "CAD",
    "ZCHF": "CHF",
    "ZEUR": "EUR",
    "ZGBP": "GBP",
    "ZJPY": "JPY",
    "ZKRW": "KRW",
    "ZUSD": "USD",
}


USUAL_TO_KRAKEN_CODEBOOK = {v: k for k, v in KRAKEN_TO_USUAL_CODEBOOK.items()}


def convert_ususal_to_kraken_code(code):
    return USUAL_TO_KRAKEN_CODEBOOK.get(code, code)


def convert_kraken_to_usual_code(code):
    return KRAKEN_TO_USUAL_CODEBOOK.get(code, code)


def kraken_download_supported_pairs():
    PAIRS_URL = "https://api.kraken.com/0/public/AssetPairs"
    resp = requests.get(PAIRS_URL)
    pairs = []
    result = resp.json()["result"]
    for _, pinfo in result.items():
        base = convert_kraken_to_usual_code(pinfo["base"])
        quote = convert_kraken_to_usual_code(pinfo["quote"])
        pairs.append(TradingPair(base=base, quote=quote))
    return list(set(pairs))


def kraken_download_next_trade_price(crypto, dtime):
    TRADES_URL = "https://api.kraken.com/0/public/Trades"
    pair = crypto + "EUR"
    since = int(dtime.timestamp())
    params = {"pair": pair, "since": since}
    logger.info(f"Downloading prices from {TRADES_URL}, parameters: {params}")
    resp = requests.get(TRADES_URL, params=params)
    result = resp.json()["result"]
    trades = list(result.items())[0][1]
    next_trade = trades[0]
    price = next_trade[0]
    return float(price)


class KrakenNextTradePriceDownloader(CachedPriceDownloader):
    def __init__(self):
        super().__init__()
        self._supported_crypto_list = self._download_supported_crypto_list()

    @property
    def supported_crypto_list(self):
        return self._supported_crypto_list

    @staticmethod
    def _download_supported_crypto_list():
        pairs = kraken_download_supported_pairs()
        return sorted(
            [p.base for p in pairs if p.quote == "EUR" and not is_fiat_currency(p.base)]
        )

    def download_price(self, crypto, dtime):
        cached_price = self.find_price_in_cache(crypto, dtime)
        if cached_price is None:
            self._download_price_add_to_cache(crypto, dtime)
        cached_price = self.find_price_in_cache(crypto, dtime)
        if cached_price is None:
            raise RuntimeError(f"Could not download price for {crypto} at {dtime}")
        return cached_price

    def _download_price_add_to_cache(self, crypto, dtime):
        price = kraken_download_next_trade_price(crypto, dtime)
        self._add_price_to_cache(crypto, dtime, price)


class MultiSourceFirstPriceDownloader(PriceDownloader):
    def __init__(self, price_downloaders):
        self.price_downloaders = price_downloaders
        self._supported_crypto_list = self._make_supported_crypto_list()

    @property
    def supported_crypto_list(self):
        return self._supported_crypto_list

    def _make_supported_crypto_list(self):
        supported_cryptos = set()
        for source in self.price_downloaders:
            source_pairs = source.supported_crypto_list
            print(source_pairs)
            supported_cryptos.update(source_pairs)
        return sorted(list(supported_cryptos))

    def download_price(self, crypto, dtime):
        for source in self.price_downloaders:
            if crypto in source.supported_crypto_list:
                try:
                    price = source.download_price(crypto, dtime)
                    return price
                except:
                    pass
        raise RuntimeError(f"Could not download price for {crypto} at {dtime}")


__price_downloader = threading.local()


def reference_price_downloader():
    price_downloader = None
    try:
        price_downloader = __price_downloader.instance
    except AttributeError:
        __price_downloader.instance = instantiate_reference_price_downloader()
        price_downloader = __price_downloader.instance
    return price_downloader


def instantiate_reference_price_downloader():
    bstamp = BitstampMinuteClosePriceDownloader()
    kraken = KrakenNextTradePriceDownloader()
    multi = MultiSourceFirstPriceDownloader([bstamp, kraken])
    return multi



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['coin2086']

package_data = \
{'': ['*']}

install_requires = \
['pandas>=1.1,<2.0', 'requests>=2.10,<3.0']

setup_kwargs = {
    'name': 'coin2086',
    'version': '0.1.3',
    'description': 'French Crypto Taxes Made Easy',
    'long_description': '# Coin2086 #\n\n> Crypto Taxes Made Easy ! 📒\n\n[![PyPI Version][pypi-image]][pypi-url]\n[![PyPI Python Version][pypi-version-image]][pypi-url]\n[![PyPI License][pypi-license-image]][pypi-url]\n[![Documentation Status][rtd-image]](https://coin2086.readthedocs.io/en/latest/?badge=latest)\n[![Binder][binder-img]][binder-url]\n\nCoin2086 is a Python module that makes it easy for French tax residents to\nfill their crypto-currency tax return.\n\nTax autorities requires French tax residents to report their profit and losses\n(PnL) on each of their cryto-currency sales of the previous year on\n[Formulaire n°2086][form2086-url].\n\nThe formula to compute your profit and losses, detailed on [Formulaire n°2086][form2086-url],\nrequires you to valuate your *whole* crypto-currency portfolio every time you sell,\nand keep track of the amount of initial investment capital that was sold. This\naccounting is tedious to do by hand.\n\nCoin2086 does all of that automatically for you. It takes your trades as input,\nvaluates your cryptocurrency portfolio, computes your taxable profit and outputs\nthe *exact* information you need to fill on [Formulaire n°2086][form2086-url].\n\nIt\'s that simple !\n\n📖  Documentation: https://coin2086.readthedocs.io/  \n📦  PyPI Package: https://pypi.org/project/coin2086/  \n📝  Example Jupyter Notebook: [Launch on Binder][binder-url]  \n💻  GitHub Project: https://github.com/fandre90/coin2086  \n\n[form2086-url]: https://www.impots.gouv.fr/portail/formulaire/2086/declaration-des-plus-ou-moins-values-de-cessions-dactifs-numeriques\n[binder-img]: https://mybinder.org/badge_logo.svg\n[binder-url]: https://mybinder.org/v2/gh/fandre90/coin2086/HEAD?filepath=notebooks%2FCoin2086%20Example%20Use.ipynb\n[pypi-image]: https://img.shields.io/pypi/v/coin2086\n[pypi-version-image]: https://img.shields.io/pypi/pyversions/coin2086\n[pypi-license-image]: https://img.shields.io/pypi/l/coin2086\n[pypi-url]: https://pypi.org/project/coin2086/\n[rtd-image]: https://readthedocs.org/projects/coin2086/badge/?version=latest\n\n## Installation ##\n\n```sh\npip install coin2086\n```\n\nAlternatively, you may use the [Binder Notebook][binder-url] directly in your browser\n\n## Basic Usage ##\n\n```python\n>>> import pandas as pd\n>>> import coin2086\n>>> trades = pd.read_csv(\'trades.csv\')\n>>> trades\n             datetime trade_side cryptocurrency  quantity     price base_currency      amount        fee\n0 2019-10-19 11:10:00        BUY            BTC      1.00   7149.38           EUR   7149.3800  35.746900\n1 2019-11-14 19:50:00       SELL            BTC      0.50   7844.88           EUR   3922.4400  19.612200\n2 2020-07-28 10:20:00        BUY            BTC      2.00   9262.42           EUR  18524.8400  92.624200\n3 2020-09-01 12:20:00        BUY            ETH      5.00    393.58           EUR   1967.9000   9.839500\n4 2020-09-05 16:50:00       SELL            BTC      1.00   8722.70           EUR   8722.7000  43.613500\n5 2020-09-08 12:40:00       SELL            ETH      5.00    285.07           EUR   1425.3500   7.126750\n6 2020-12-20 09:10:00       SELL            BTC      0.25  19223.90           EUR   4805.9750  24.029875\n7 2021-03-13 23:40:00       SELL            BTC      0.25  50025.17           EUR  12506.2925  62.531463\n\n>>> year = 2020\n>>> form2086, taxable_profit = coin2086.compute_taxable_pnls(trades, year=year)\n>>> print(f"Total taxable profit for year {year}: {taxable_profit:.2f} euros")\nTotal taxable profit for year 2020: 2038.50 euros\n>>> form2086\n     Description  ... Plus-values et moins-values [pnl]\n4  SELL 1.00 BTC  ...                       -371.708792\n5  SELL 5.00 ETH  ...                       -102.332358\n6  SELL 0.25 BTC  ...                       2512.542417\n\n[3 rows x 10 columns]\n```\n\nFor more information, check out the \n[documentation](https://coin2086.readthedocs.io/) or the\n[example notebook][binder-url]\n',
    'author': 'Fabien André',
    'author_email': 'fabien.andre90@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/fandre90/coin2086',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.2,<4.0',
}


setup(**setup_kwargs)



import pandas as pd

SUPPORTED_CRYPTOS = [
    "AAVE",
    "BAT",
    "BCH",
    "BTC",
    "ETH",
    "KNC",
    "LINK",
    "LTC",
    "MKR",
    "OMG",
    "PAX",
    "SNX",
    "UMA",
    "UNI",
    "USDC",
    "XLM",
    "XRP",
    "YFI",
    "ZRX",
    "ADA",
    "ALGO",
    "ANT",
    "ATOM",
    "BAL",
    "COMP",
    "CRV",
    "DAI",
    "DASH",
    "DOGE",
    "DOT",
    "EOS",
    "ETC",
    "EWT",
    "FIL",
    "FLOW",
    "GNO",
    "GRT",
    "ICX",
    "KAVA",
    "KEEP",
    "KSM",
    "LSK",
    "MANA",
    "MLN",
    "NANO",
    "OCEAN",
    "OXT",
    "PAXG",
    "QTUM",
    "REP",
    "REPV2",
    "SC",
    "STORJ",
    "TBTC",
    "TRX",
    "USDT",
    "WAVES",
    "XMR",
    "XTZ",
    "ZEC",
]


def check_trades(trades):
    mandatory_columns = set(
        [
            "datetime",
            "trade_side",
            "cryptocurrency",
            "quantity",
            "price",
            "base_currency",
            "amount",
            "fee",
        ]
    )
    missing = mandatory_columns - set(trades.columns)
    if len(missing) > 0:
        raise ValueError(f"Missing columns from trades dataframe {missing}")
    currencies = set(trades["base_currency"].drop_duplicates())
    if currencies != set(["EUR"]):
        raise ValueError("Base currency (base_currency) must be EUR for all trades")
    sides = set(trades["trade_side"].drop_duplicates())
    if sides != set(["SELL", "BUY"]):
        raise ValueError("Trade side (trade_side) must be either BUY or SELL")
    sorted_supported = ",".join(sorted(SUPPORTED_CRYPTOS))
    supported = set(SUPPORTED_CRYPTOS)
    cryptos = set(trades["cryptocurrency"].drop_duplicates())
    unsupported = cryptos - supported
    if len(unsupported) > 0:
        unsupported_sorted = ",".join(sorted(list(unsupported)))
        raise ValueError(
            f"Unsupported cryptocurrencies: {unsupported_sorted} "
            + f"supported currencies are: {sorted_supported}"
        )
    unsigned_cols = ["quantity", "price", "amount", "fee"]
    subset = trades[unsigned_cols]
    any_negative = (subset < 0).any(axis=None)
    if any_negative:
        cols = ",".join(unsigned_cols)
        raise ValueError(
            f"The columns {cols} are unsigned. All values MUST be positive."
        )
    sorted_trades = trades.sort_values("datetime").reset_index()
    sorted_trades = sorted_trades.drop(columns="index")
    if not sorted_trades.equals(trades):
        raise ValueError(
            f"It looks like your trades are not sorted by increasing datetime "
            f"with a monotic index. This can usually be fixed with "
            f"trades.sort_values('datetime').reset_index().drop(columns='index')"
        )
    trades["datetime"] = pd.to_datetime(trades["datetime"])
